from pinjected.di.injected import Injected
from pinjected.di.decorators import injected_function, injected_instance, injected, instance, reload
from pinjected.di.util import EmptyDesign, instances, providers, classes
from pinjected.di.design import Design
from pinjected.di.designed import Designed

__all__ = [
    "Injected",
    "Design",
    "EmptyDesign",
    "instances",
    "providers",
    "classes",
    "instance",
    "injected",
    "reload"
]



import glob
from pathlib import Path
from typing import List

import loguru

from pinjected import instances, Injected, injected, providers

__meta_design__ = instances(
    default_design_paths=[
        "davinci.paper_indexer_design.gpt4_llm_design"
    ],
    #default_design_paths=["davinci.paper_indexer_design.gpt35_llm_design"],
    overrides=providers(
        logger=lambda: loguru.logger
    )
)

from pinjected.llm_support.inspect_module_prompts import INSPECT_TEMPLATE

def test(*xyz):
    pass

@injected
def count_tokens(count_token, /, targets: List[Path]):
    # find .py files recursively
    # count tokens in each file
    for pyf in targets:
        token = count_token(Path(pyf).read_text())
        print(f"file:{pyf} -> tokens:{token}")


@injected
def list_python_files(root: Path) -> List[Path]:
    # find .py files recursively
    # count tokens in each file
    py_files = list(glob.glob(str(root / '**/*.py'), recursive=True))
    return [Path(pyf) for pyf in py_files]


all_py_files = list_python_files(Path("./"))
see_tokens: Injected = count_tokens(list_python_files(Path("./")))


@injected
async def inspect_type(a_cached_smart_llm: 'str -> str', /, py_text: str) -> str:
    prompt = INSPECT_TEMPLATE + f"""
Now, please answer the infered type for each global declarations in the following python script.
```python
{py_text}
```
If the type cannot be infered, please answer 'unknown'.
For callable type signature, use T->U to represent a function that takes a T and returns a U, instead of Callable[[T],U].
"""
    return await a_cached_smart_llm(prompt)



@injected
def test_function(smart_prompt_helper,/,x):
    print(smart_prompt_helper)
    return x


ans:Injected = inspect_type(Path("/Users/s22625/repos/pinject-design/pinjected/llm_support/inspect_module.py").read_text())



from functools import partial
from typing import Callable, Any

from loguru import logger

from pinjected.di.injected import Injected
from pinjected.di.util import instances
from pinjected import Design, injected_function


def _factory(a, b, y=0, x=7):
    assert a == 3
    assert b == 2
    assert x == 5
    return a + b + x


def test_partial():
    d = Design().bind_instance(
        a=1,
        b=2
    ).bind_provider(
        f=Injected.partial(_factory, b=Injected.by_name("b"), x=Injected.pure(5))
    )
    assert d.provide("f")(3) == 10


def test_injected_function():
    @injected_function
    def test_func(_x, _y, /, *args, **mykwargs):
        assert args, "args should be non-empty"
        return _x + _y + str(args)

    g = instances(
        x='x',
        y='y',
    ).to_graph()
    assert g[test_func](1, 2, 3, 5, 6) == 'x' + 'y' + '(1, 2, 3, 5, 6)'


def test_injected_function_with_defaults():
    @injected_function
    def test_func(x, y, /, a, b, *args, z=7, **mykwargs):
        return x, y, (a, b, *args), mykwargs, z

    @injected_function
    def test_func2(_x,_y, a, b, z=7,s=10):
        return _x, _y, (a, b), z,s


    g = instances(
        x='x',
        y='y',
    ).to_graph()
    expectation = ("x", 'y', (1, 2, 3, 5, 6), {}, 7)
    expectation2 = ("x", 'y', (1, 2, 3, 5, 6), {}, 9)
    expectation3 = ("x", 'y', (1, 2, 3, 5, 6), {'p': 42}, 7)
    expectation4 = ("x", 'y', (1, 2, 3, 5, 6), {'p': 42}, 9)
    expectation5 = ("x", 'y', (1, 2), {}, 7)
    expectation6 = ("x", 'y', (1, 2), 7,10)
    assert g[test_func(1, 2, 3, 5, 6)] == expectation
    assert g[test_func](1, 2, 3, 5, 6) == expectation
    assert g[test_func](1, 2, 3, 5, 6, z=9) == expectation2
    assert g[test_func(1, 2, 3, 5, 6, z=9)] == expectation2
    assert g[test_func(1, 2, 3, 5, 6, p=42)] == expectation3
    assert g[test_func](1, 2, 3, 5, 6, p=42) == expectation3
    assert g[test_func](1, 2, 3, 5, 6, p=42, z=9) == expectation4
    assert g[test_func(1, 2, 3, 5, 6, p=42, z=9)] == expectation4
    assert g[test_func(1, 2)] == expectation5
    assert g[test_func](1, 2) == expectation5
    assert g[test_func2](1, 2) == expectation6
    assert g[test_func2(1, 2)] == expectation6
    assert g[test_func2(1, 2, 7)] == expectation6




from abc import ABC, abstractmethod
from typing import Generic

from pinjected.di.proxiable import T


class Applicative(Generic[T], ABC):
    @abstractmethod
    def map(self, target: T, f) -> T:
        pass

    @abstractmethod
    def zip(self, *targets: T):
        pass

    @abstractmethod
    def pure(self, item) -> T:
        pass

    @abstractmethod
    def is_instance(self, item)->bool:
        pass

    def dict(self, **kwargs: T) -> T:
        items = list(kwargs.items())
        keys = [t[0] for t in items]
        values = [t[1] for t in items]
        return self.zip(*values).map(lambda vs: dict(zip(keys, vs)))



from dataclasses import dataclass
from pathlib import Path
from typing import Union

from pinjected.module_var_path import ModuleVarPath


@dataclass
class ModuleVarLocation:
    path: Path
    line: int
    column: int

CodeLocation = Union[
    ModuleVarPath,
    ModuleVarLocation
]
"""
a class to indicate the location of a python variable.
1. qualified_name i.e. ModuleVarPath: the qualified name of the variable.
2. file_path + line + column: the line and column of the variable.
"""



import ast
import json
from contextlib import contextmanager
from pathlib import Path
from pprint import pformat

import click

from pinjected import injected, Injected


@click.group()
def main():
    pass


@main.command()
@click.argument('script_path', type=click.Path(exists=True))
@click.argument('func_name')
def extract_func_source_and_imports(script_path: str, func_name: str):
    func_and_imports = extract_func_source_and_imports_dict(func_name, script_path)

    # Prepare the data containing the requested function source and imports
    data = {
        'code': func_and_imports['functions'].get(func_name, f"Function {func_name} not found."),
        'imports': list(set(func_and_imports['imports']))
    }
    # Convert the data to JSON format and print it
    text = json.dumps(data, indent=4)
    # Ensure that the output does not contain "<pinjected>" to prevent code injection
    assert "<pinjected>" not in text
    text = "<pinjected>\n" + text + "\n</pinjected>"
    print(text)


def extract_func_source_and_imports_dict(script_path):
    # Read the content of the script file
    file_content = Path(script_path).read_text()
    # Parse the file content into an AST node tree
    node_tree = ast.parse(file_content)

    # Function to get the source code of decorators with "@"
    def get_decorators_source(decorators):
        return ["@" + ast.get_source_segment(file_content, decorator).strip() for decorator in decorators]

    # Initialize a dictionary to store functions and imports
    func_and_imports = {'functions': {}, 'imports': []}
    # Walk through the AST nodes
    for node in ast.walk(node_tree):
        # Check if the node is a function
        if isinstance(node, ast.FunctionDef):
            # Get decorators source code with "@"
            decorators_source = get_decorators_source(node.decorator_list)
            # Get the source code of the function definition
            func_source_lines = file_content.splitlines()[node.lineno - 1:node.end_lineno]
            func_source = "\n".join(decorators_source + func_source_lines)
            # Store the function source code in the dictionary
            func_and_imports['functions'][node.name] = func_source
        # Check if the node is an import or import from statement
        elif isinstance(node, (ast.Import, ast.ImportFrom)):
            # Get the source code of the import
            import_source = ast.get_source_segment(file_content, node)
            # Store the import source code in the list
            func_and_imports['imports'].append(import_source)
    return func_and_imports


def create_run_design(script_path: str):
    """
    This command should load the __meta_design__ and return the final design.
    :param script_path:
    :return:
    """
    from pinjected.helper_structure import MetaContext
    __design__ = MetaContext.gather_from_path(Path(script_path)).accumulated
    __graph__ = __design__.to_graph().auto_sync()
    # mc:MetaContext = MetaContext.gather_from_path(Path(script_path))
    # return mc.accumulated


class WithBlockVisitor(ast.NodeVisitor):
    def __init__(self):
        self.result = {}

    def visit_With(self, node):
        # In Python 3.9 and above, 'with' statements use 'items' which is a list of 'withitem' objects.
        # In earlier versions, the context expression was directly available as 'context_expr'.
        # We'll check for both to ensure compatibility.
        with_items = getattr(node, 'items', None) or [node]

        for item in with_items:
            context_expr = getattr(item, 'context_expr', None) or item

            # Check if the with statement contains a call to reload
            if isinstance(context_expr, ast.Call) and getattr(context_expr.func, 'id', None) == 'reload':
                # Get the list of reload targets
                reload_targets = [arg.s for arg in context_expr.args]
                # Process each item in the with block
                for body_item in node.body:
                    if isinstance(body_item, ast.AnnAssign) and isinstance(body_item.target, ast.Name):
                        variable_name = body_item.target.id
                        # Assign reload targets to the variable in the result dict
                        self.result[variable_name] = {
                            'reload_targets': reload_targets
                        }


def extract_with_block_structure(source_code):
    # Parse the source code into an AST
    parsed_ast = ast.parse(source_code)
    # Create a visitor instance
    visitor = WithBlockVisitor()
    # Visit the parsed AST
    visitor.visit(parsed_ast)
    # Return the collected results
    return visitor.result


def extrract_assignments(source_code):
    """
    Extract assignment statements from the given Python source code.
    Returns a dictionary with variable names as keys and the source code of the assigned value as values.
    This updated version handles type annotations.
    """

    class AssignmentVisitor(ast.NodeVisitor):
        def __init__(self):
            self.assignments = {}

        def visit_AnnAssign(self, node):
            # Handle assignments with annotations
            if isinstance(node.target, ast.Name):
                var_name = node.target.id
                value_code = ast.get_source_segment(source_code, node.value)
                self.assignments[var_name] = value_code

        def visit_Assign(self, node):
            # Handle normal assignments
            for target in node.targets:
                if isinstance(target, ast.Name):
                    var_name = target.id
                    value_code = ast.get_source_segment(source_code, node.value)
                    self.assignments[var_name] = value_code
                # Handle tuple assignments
                elif isinstance(target, ast.Tuple):
                    for elt in target.elts:
                        if isinstance(elt, ast.Name):
                            var_name = elt.id
                            value_code = ast.get_source_segment(source_code, node.value)
                            self.assignments[var_name] = value_code

    # Parse the source code into an AST
    parsed_ast = ast.parse(source_code)
    # Create a visitor instance
    visitor = AssignmentVisitor()
    # Visit the parsed AST
    visitor.visit(parsed_ast)
    # Return the collected assignments
    return visitor.assignments

@main.command()
@click.argument('script_path', type=click.Path(exists=True))
@click.argument('target_name')
def generate_code_with_reload(
        script_path: str,
        target_name: str,
):
    """
    1. find all function names using extract_func_source_and_imports
    2. check the reload target of target_name
    3. concat all sources and return it as a pinjected json
    :param script_path:
    :param target_name:
    :return:
    """
    from loguru import logger
    source = Path(script_path).read_text()
    func_table = extract_func_source_and_imports_dict(script_path)
    logger.info(f"func_table:{pformat(func_table)}")
    reload_table = extract_with_block_structure(source)
    if target_name not in reload_table:
        reload_targets = []
    else:
        reload_targets = reload_table[target_name]['reload_targets']
    funcs_to_reload = [func_table['functions'][func_name] for func_name in reload_targets]
    imports = func_table['imports']
    imports = '\n'.join(imports)
    func_defs = '\n'.join(funcs_to_reload)
    assignments: dict[str, str] = extrract_assignments(source)
    # logger.info(pformat(assignments))

    """
    How can I add the child session?
    """
    code = f"""
if '__graph__' not in globals():
    from pinjected.helper_structure import MetaContext
    from pathlib import Path
    exec(Path("{script_path}").read_text())
    __meta_context__ = MetaContext.gather_from_path(Path("{script_path}"))
    __design__ = __meta_context__.final_design
    __graph__ = __design__.to_graph().auto_sync()
{imports}
{func_defs}
{target_name} = {assignments[target_name]}
__graph__[{target_name}]
"""
    # oops, target_name is not in the scope
    logger.debug(f"generated code:{code}")
    data = "<pinjected>\n" + json.dumps({'code': code}) + "\n</pinjected>"
    print(data)


@injected
def test_target_function():
    print(f"hello")


@contextmanager
def reload(*targets):
    # this is just a placeholder for the AST visitor
    yield


with reload('test_target_function'):
    x = 0
    run_test: Injected = test_target_function


def is_pydevd():
    import os
    return 'PYDEVD_LOAD_VALUES_ASYNC' in os.environ


if __name__ == '__main__' and not is_pydevd():
    """
    this code generation works, 
    now the remaining task is, 
    how to get the script path and its var name,
    
    Options:
    1. use the gutter icon
    2. action: run current thing? nah. 
    so,, what is the way to rerun the current thing? quickly?
    I think, placing a cursor on the target and then running the command is the best way.
    so, let's make an action. 
    """
    main()



import inspect
from functools import wraps
from pathlib import Path

from pprint import pformat

import asyncio

from typing import Coroutine, Awaitable

from loguru import logger
from returns.result import safe

from pinjected import instances, Injected, Design, providers
from pinjected.helper_structure import MetaContext
from pinjected.helpers import get_design_path_from_var_path
from pinjected.module_var_path import ModuleVarPath, load_variable_by_module_path
from pinjected.logging_helper import disable_internal_logging
from pinjected.notification import notify


def run_injected(
        cmd,
        var_path,
        design_path: str = None,
        *args,
        **kwargs
):
    logger.info(
        f"run_injected called with cmd:{cmd}, var_path:{var_path}, design_path:{design_path}, args:{args}, kwargs:{kwargs}")
    # TODO refactor to merge run_injected and run_anything
    with disable_internal_logging():
        if "return_result" in kwargs:
            return_result = kwargs.pop("return_result")
        else:
            return_result = False
        if "overrides" in kwargs:
            overrides = kwargs.pop("overrides")
        else:
            overrides = instances()
        if design_path is None:
            design_path = get_design_path_from_var_path(var_path)
        logger.info(
            f"run_injected called with cmd:{cmd}, var_path:{var_path}, design_path:{design_path}, args:{args}, kwargs:{kwargs}")
    return run_anything(cmd,
                        var_path,
                        design_path=design_path,
                        return_result=return_result,
                        overrides=overrides,
                        call_args=args,
                        call_kwargs=kwargs,
                        )


def run_anything(
        cmd: str,
        var_path,
        design_path,
        overrides=instances(),
        return_result=False,
        call_args=None,
        call_kwargs=None,
        notify=lambda msg, *args, **kwargs: notify(msg, *args, **kwargs)
):
    from loguru import logger
    with disable_internal_logging():

        loaded_var = load_variable_by_module_path(var_path)
        meta = safe(getattr)(loaded_var, "__runnable_metadata__").value_or({})
        if not isinstance(meta, dict):
            meta = {}
        overrides += meta.get("overrides", instances())

        meta_cxt: MetaContext = MetaContext.gather_from_path(ModuleVarPath(var_path).module_file_path)
        if design_path is None:
            design_path = meta_cxt.accumulated.provide("default_design_paths")[0]

        var: Injected = Injected.ensure_injected(load_variable_by_module_path(var_path))
        design: Design = load_variable_by_module_path(design_path)
        meta_design = instances(overrides=instances()) + meta_cxt.accumulated
        overrides += meta_design.provide("overrides")

    design = design + overrides
    logger.info(f"running target:{var} with {design_path} + {overrides}")
    logger.debug(design.keys())
    # logger.info(f"running target:{var} with cmd {cmd}, args {args}, kwargs {kwargs}")
    # logger.info(f"metadata obtained from pinjected: {meta}")


    res = None
    try:
        if cmd == 'call':
            args = call_args or []
            kwargs = call_kwargs or {}
            logger.info(f"run_injected call with args:{args}, kwargs:{kwargs}")
            res = design.provide(var)(*args, **kwargs)
            if isinstance(res, Coroutine):
                res = asyncio.run(res)
            logger.info(f"run_injected call result:\n{res}")
            if isinstance(res, Awaitable):
                async def impl():
                    return await res
                logger.info(f"awaiting awaitable")
                res = asyncio.run(impl())
        elif cmd == 'get':
            logger.info(f"providing...")
            res = design.provide(var)
            if isinstance(res, Coroutine):
                res = asyncio.run(res)
            if isinstance(res, Awaitable):
                async def impl():
                    return await res
                logger.info(f"awaiting awaitable")
                res = asyncio.run(impl())
            if not return_result:
                logger.info(f"run_injected get result:\n{pformat(res)}")
        elif cmd == 'fire':
            return_result = True
            res = design.provide(var)
            if isinstance(res, Coroutine):
                res = asyncio.run(res)
            logger.info(f"run_injected fire result:\n{res}")
            if inspect.iscoroutinefunction(res) or (hasattr(res,'__is_async__') and res.__is_async__):
                logger.info(f'{res} is a coroutine function, wrapping it with asyncio.run')
                src = res
                #@wraps(res)
                def synced(*args, **kwargs):
                    return asyncio.run(src(*args, **kwargs))
                res = synced
            else:
                logger.info(f"{res} is not a coroutine function.")
        elif cmd == 'visualize':
            from loguru import logger
            logger.info(f"visualizing {var_path} with design {design_path}")
            logger.info(f"deps:{var.dependencies()}")
            design.to_vis_graph().show_injected_html(var)
        elif cmd == 'to_script':
            from loguru import logger
            d = design + providers(
                __root__=var
            )
            print(d.to_vis_graph().to_python_script(var_path, design_path=design_path))
    except Exception as e:
        import traceback
        notify(f"Run failed with error:\n{e}", sound='Frog')
        trace = traceback.format_exc()
        Path(f"run_failed_{var_path}.err.log").write_text(str(e) + "\n" + trace)
        # from rich.console import Console
        # console = Console()
        # console.print_exception(show_locals=False)
        raise e
    notify(f"Run result:\n{str(res)[:100]}")
    if return_result:
        return res



import inspect
import re
from dataclasses import dataclass
from pprint import pformat

from loguru import logger
from pinjected import Injected
from pinjected.di.bindings import Bind, InjectedBind


def default_get_arg_names_from_class_name(class_name):
    """Converts normal class names into normal arg names.

    from github's pinject

    Normal class names are assumed to be CamelCase with an optional leading
    underscore.  Normal arg names are assumed to be lower_with_underscores.

    Args:
      class_name: a class name, e.g., "FooBar" or "_FooBar"
    Returns:
      all likely corresponding arg names, e.g., ["foo_bar"]
    """
    parts = []
    rest = class_name
    if rest.startswith('_'):
        rest = rest[1:]
    while True:
        m = re.match(r'([A-Z][a-z]+)(.*)', rest)
        if m is None:
            break
        parts.append(m.group(1))
        rest = m.group(2)
    if not parts:
        return []
    return ['_'.join(part.lower() for part in parts)]


def find_classes(modules, classes):
    if classes is not None:
        all_classes = set(classes)
    else:
        all_classes = set()
    for module in _get_explicit_or_default_modules(modules):
        # TODO(kurts): how is a module getting to be None??
        if module is not None:
            all_classes |= _find_classes_in_module(module)
    return all_classes


def _get_explicit_or_default_modules(modules):
    if modules is None:
        return []
    return modules


def _find_classes_in_module(module):
    classes = set()
    for member_name, member in inspect.getmembers(module):
        if inspect.isclass(member) and not member_name == '__class__':
            classes.add(member)
    return classes


@dataclass
class DIGraphHelper:
    src: "Design"

    def get_implicit_mapping(self) -> dict[str, type]:
        classes = find_classes(self.src.modules, self.src.classes)
        for c in classes:
            for name in default_get_arg_names_from_class_name(c.__name__):
                yield name, c

    def get_explicit_mapping(self) -> dict[str, Bind]:
        return {k: b for k, b in self.src.bindings.items()}

    def total_mappings(self) -> dict[str, Injected]:
        return {k: v.to_injected() for k, v in self.total_bindings().items()}

    def total_bindings(self) -> dict[str, Bind]:
        from pinjected.di.implicit_globals import IMPLICIT_BINDINGS
        global_implicit_mappings = IMPLICIT_BINDINGS
        global_implicit_mappings = {k: v for k, v in global_implicit_mappings.items()}
        # TODO add the qualified name for the global_implicit_mappings. but how?

        # logger.debug(f"global_implicit_mappings: {pformat(global_implicit_mappings)}")
        logger.debug(f"using {len(global_implicit_mappings)} global implicit mappings")
        implicit_mappings = {k: InjectedBind(Injected.bind(v)) for k, v in self.get_implicit_mapping()}
        explicit_mappings = self.get_explicit_mapping()
        return {**global_implicit_mappings, **implicit_mappings, **explicit_mappings}




from abc import abstractmethod
from dataclasses import dataclass
from typing import Generic, TypeVar, Any

T = TypeVar("T")


class IProxyContext(Generic[T]):

    @abstractmethod
    def getattr(self, tgt: T, name: str):
        pass

    @abstractmethod
    def call(self, tgt: T, *args, **kwargs):
        pass

    @abstractmethod
    def pure(self, tgt: T) -> "DelegatedVar[T]":
        pass

    @abstractmethod
    def getitem(self, tgt: T, key) -> Any:
        pass

    @abstractmethod
    def eval(self, tgt: T):
        pass

    @abstractmethod
    def alias_name(self):
        pass

    @abstractmethod
    def iter(self, tgt: T):
        pass

    @abstractmethod
    def dir(self, tgt):
        pass

    def biop_impl(self, op: str, tgt: T, other):
        """Bi-Operator Implementation
        op: '+' | '-' | '*' | '/' | '%' | '**' | '<<' | '>>' | '&' | '^' | '|' | '//'
        """
        raise NotImplementedError()

    def unary_impl(self, op: str, tgt: T):
        """Unary Operator Implementation
        op: '-' | '~' | 'len' | 'del'
        """
        raise NotImplementedError()


@dataclass
class DelegatedVar(Generic[T]):
    # Generic Var Type that delegates all implementations to cxt.
    value: T
    cxt: IProxyContext[T]

    def __getattr__(self, item):
        return self.cxt.getattr(self.value, item)

    def __call__(self, *args, **kwargs):
        return self.cxt.call(self.value, *args, **kwargs)

    def __getitem__(self, key):
        return self.cxt.getitem(self.value, key)

    def eval(self) -> T:
        """this should return the semantic result of calculation."""
        return self.cxt.eval(self.value)

    def __str__(self):
        return f"{self.cxt.alias_name()}({self.value},{self.cxt})"

    def __iter__(self):
        return self.cxt.iter(self.value)

    def __getstate__(self):
        return self.value, self.cxt

    def __setstate__(self, state):
        self.value, self.cxt = state

    def __hash__(self):
        return hash((self.value, self.cxt))

    def __add__(self, other):
        return self.cxt.biop_impl('+', self.value, other)

    def __mul__(self, other):
        return self.cxt.biop_impl('*', self.value, other)

    def __truediv__(self, other):
        return self.cxt.biop_impl('/', self.value, other)

    def __mod__(self, other):
        return self.cxt.biop_impl('%', self.value, other)



import ast
import importlib
import sys
from dataclasses import dataclass
from pathlib import Path


@dataclass
class ModuleVarPath:
    """
    represents a path where a variable is defined.
    like a.b.c.d
    """
    path: str

    def __post_init__(self):
        assert self.module_name is not None
        assert self.var_name is not None

    def load(self):
        return load_variable_by_module_path(self.path)

    @property
    def module_name(self):
        module = ".".join(self.path.split(".")[:-1])
        return module

    @property
    def var_name(self):
        return self.path.split(".")[-1]

    def to_import_line(self):
        return f"from {self.module_name} import {self.var_name}"

    @property
    def module_file_path(self):
        return sys.modules[self.module_name].__file__

    @staticmethod
    def from_local_variable(var_name):
        # get parent frame
        # then return a ModulePath
        raise NotImplementedError

    def definition_snippet(self):
        """
        returns a code snippet that defines the variable
        :return:
        """
        return find_var_or_func_definition_code_in_module(self.module_name, self.var_name)

    def depending_import_lines(self):
        return find_import_statements_in_module(self.module_name)


def load_variable_by_module_path(full_module_path):
    from loguru import logger
    logger.info(f"loading {full_module_path}")
    module_path_parts = full_module_path.split('.')
    variable_name = module_path_parts[-1]
    module_path = '.'.join(module_path_parts[:-1])

    # Import the module
    module = importlib.import_module(module_path)

    # Retrieve the variable using getattr()
    if not hasattr(module, variable_name):
        logger.warning(f"Module {module_path} at {module.__file__} does not have variable {variable_name}. src = \n{Path(module.__file__).read_text()}")
        raise RuntimeError(f"Module {module_path} at {module.__file__} does not have variable {variable_name}. but has {dir(module)}")
    variable = getattr(module, variable_name)

    logger.info(f"loaded {full_module_path}")

    return variable


def find_var_or_func_definition_code_in_module(module_dot_path, name):
    try:
        module = importlib.import_module(module_dot_path)
    except ImportError:
        return f"Module {module_dot_path} could not be imported."

    module_file_path = module.__file__

    if module_file_path.endswith('.pyc'):
        module_file_path = module_file_path[:-1]  # convert .pyc to .py

    with open(module_file_path, 'r') as f:
        lines = f.readlines()
        tree = ast.parse(''.join(lines))

    definition = None

    for node in ast.walk(tree):
        if definition is None and isinstance(node, (ast.Assign, ast.AnnAssign, ast.FunctionDef,ast.AsyncFunctionDef)):
            targets = []
            if isinstance(node, ast.Assign):
                targets = node.targets
            elif isinstance(node, ast.AnnAssign):
                targets = [node.target]

            for target in targets:
                if isinstance(target, ast.Name) and target.id == name:
                    start_line = node.lineno  # 1-based index
                    end_line = getattr(node, 'end_lineno', start_line)  # Some older Python versions don't have end_lineno
                    definition = ''.join(lines[start_line - 1:end_line]).strip()

            if isinstance(node, (ast.FunctionDef,ast.AsyncFunctionDef)) and node.name == name:
                start_line = node.lineno
                end_line = getattr(node, 'end_lineno', start_line)
                definition = ''.join(lines[start_line - 1:end_line]).strip()

    if definition:
        return definition
    else:
        raise RuntimeError(f"{name} is not defined in the given module: {module_dot_path}")


def find_import_statements_in_module(module_dot_path):
    try:
        module = importlib.import_module(module_dot_path)
    except ImportError:
        return f"Module {module_dot_path} could not be imported."

    module_file_path = module.__file__

    if module_file_path.endswith('.pyc'):
        module_file_path = module_file_path[:-1]  # convert .pyc to .py

    with open(module_file_path, 'r') as f:
        lines = f.readlines()
        tree = ast.parse(''.join(lines))

    import_statements = []

    for node in ast.walk(tree):
        if isinstance(node, (ast.Import, ast.ImportFrom)):
            start_line = node.lineno
            end_line = getattr(node, 'end_lineno', start_line)
            import_statements.append(''.join(lines[start_line - 1:end_line]).strip())

    return import_statements
